#!/usr/bin/env janet

# Copyright (C) 2023, Maxim Likhachev, <envrm@yandex.ru>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

(import http)
(use joy)

(import joy/responder :as responder)
(import spork/json :as json)

# ------------------------------------------------------------------------------

(def frisbi-api-url "https://frisbi24.ru/api/account")

(defn usage (args)
  (print (string/join
           @((string/format "%s [--help] - https://frisbi24.ru metrics exporter.\n" (last (string/split "/" (args 0))))
             "  Configuration environment variables:"
             "    FRISBI_EXPORTER_PORT - TCP port to bind       (default: 127.0.0.1)"
             "    FRISBI_EXPORTER_HOST - IP address to listen   (default: 8000)\n"
             "  Usage:"
             "    curl 'http://host:port/metrics?account_id=ID&service_id=ID'")
           "\n")))

# ------------------------------------------------------------------------------

(defn parse-date (date)
  (let (d (map scan-number (string/split "." date)))
    {:year (d 2)
     :month (- (d 1) 1)
     :month-day (- (d 0) 1)}))

(defn date-to-timestamp (date)
  (string (os/mktime (parse-date date))))

(defn format-labels (labels)
  (string/join
    (map |(string/format "%s=\"%s\"" (first $) (last $)) (pairs labels)) ", "))

(defn get-name (data)
  (first (string/split " " (data "name"))))

(defmacro meter/metadata-value-pair (value)
  ~(let (meters (data "meters"))
     (map (fn (id)
            (let (meter (meters id))
              @{:metadata @{:id id :name (get-name meter)}
               :value   ,value}))
          (keys meters))))

(defn get-meter-expiration-date (data)
  (meter/metadata-value-pair
    (date-to-timestamp ((meter "expiration") "date"))))

(defn get-meter-values (data)
  (meter/metadata-value-pair
    (meter "value")))

(defn get-meter-states (data)
  (meter/metadata-value-pair
    (if (= ((meter "expiration") "state") "active") "1" "0")))

(defn get-debt (data)
  (keep |(let (service ($ "service"))
           (when (service "supports_invoices")
             @{:metadata @{:id (string (service "id")) :name (service "name")}
              :value    (string (/ ($ "amount") 100))}))
        (data "balance")))

(defn create-metric (description metric metric-type common-labels data)
  (defn format-metric (metric labels value)
    (string/format "%s{%s} %s" metric (format-labels labels) value))

  (array
    (string/format "\n# HELP %s %s" metric description)
    (string/format   "# TYPE %s %s" metric metric-type)
    (splice (map |(format-metric metric (merge common-labels ($ :metadata)) ($ :value)) data))))

(defn m (metric metric-type data)
  (def common-labels @{:account (data "account")
                       :address (data "address")})
  (string/join
    (array/concat
      (create-metric "Задолженности"           'frisbi_debt              'gauge   common-labels (get-debt data))
      (create-metric "Показания счётчиков"     'frisbi_meter_value       'counter common-labels (get-meter-values data))
      (create-metric "Активность счётчиков"    'frisbi_meter_state       'counter common-labels (get-meter-states data))
      (create-metric "Срок действия счётчиков" 'frisbi_meter_expiry_date 'gauge   common-labels (get-meter-expiration-date data))
      "")
    "\n"))

(defn frisbi-api (account-id service-id)
  (string/join @(frisbi-api-url account-id service-id) "/"))

(defn api (account-id service-id)
  (if (or (nil? account-id) (nil? service-id))
    (responder/respond
      :text "Query parameters 'account_id' and 'service_id' are required: e.g. GET /metrics?account_id=1&service_id=2"
      :status 400)

    (do
      (def content
        (json/decode
          (get (http/get (frisbi-api account-id service-id)
                         :headers {"Accept" "application/json"}
                         :follow-redirects true)
               :body)))

      (let (ok      (content "is_ok")
               reason  (content "description")
               payload (content "payload"))

        (if (not ok)
          (responder/respond :text reason :status (if (= reason "Not Found") 404 400))
          (responder/respond :text (m 'frisbi_meters_value 'gauge payload)))))))

# ------------------------------------------------------------------------------

(def index-page
  [:html
   [:head
    [:title "Frisbi24.ru Exporter"]]
   [:body
    [:h1 "Frisbi24.ru Exporter"]
    [:p [:a {:href "/metrics"} "Metrics"]]]])

(def default-exporter-host "127.0.0.1")
(def default-exporter-port 8000)

# ------------------------------------------------------------------------------

(defn / (request)
  (text/html index-page))

(def not-found /)

(defn /metrics (request)
  (var query (request :query-string))
  (if (nil? query)
    (set query @{}))
  (api (query :account_id) (query :service_id)))

(def app (app {:404 not-found}))

# ------------------------------------------------------------------------------

(defn main (& args)
  (when (or (= (get args 1) "-h")
                (= (get args 1) "--help"))
    (usage args)
    (os/exit))

  (var exporter-port (os/getenv "FRISBI_EXPORTER_PORT"))
  (var exporter-host (os/getenv "FRISBI_EXPORTER_HOST"))

  (when (nil? exporter-host)
    (set exporter-host default-exporter-host)
    (print (string/format "FRISBI_EXPORTER_HOST is not set. Using %s." default-exporter-host)))

  (when (nil? exporter-port)
    (set exporter-port default-exporter-port)
    (print (string/format "FRISBI_EXPORTER_PORT is not set. Using %d" default-exporter-port)))

  (server app exporter-port exporter-host))

